###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from copy import copy, deepcopy
from inspect import isroutine
from typing import Any, Self

from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import DefaultError, RequiredError, FieldValueError


def __get_field_value__(obj: Any, attr: str, value: Any) -> Any:
    """
    Function that get the cleaned value for a Field and validate this value.

    Args:
        obj (Any): A class or a instance of BaseObject.
        attr (str): The attribute name.
        value (Any): The value that is assigned to this attribute.

    Raises:
        FieldValueError: If we find validation errors.
    """
    # Transforms the given value to Undefined if it matches the default_parse_string.
    value = _transform_to_undefined(value)

    # Get all attributes that the object has
    attributes = getattr(obj, MetaClass._attr_name) # pylint: disable=protected-access
    try:
        field = attributes[attr]
    except KeyError:
        return value

    # field can be the type itself or an instance of BaseField
    if isinstance(field, BaseField):
        try:
            value = field.get_cleaned_value(value)
        except Exception as error:
            # Add attribute name to error
            error.args = (f'{attr}: {str(error)}',)
            raise FieldValueError(error.args) from error

        field.validate(attr, value)
    else:
        _validate(attr, value, field)

    return value


def _transform_to_undefined(value: Any) -> Any:
    """
    Transforms the given value to Undefined if it matches the default parse string.

    Args:
        value (Any): The value to transform.

    Returns:
        Any: The transformed value.
    """
    if isinstance(value, str) and value == Undefined.default_parse_string:
        value = Undefined

    return value


def _required(attr_name: str, value: Any) -> None:
    """
    Checks if value is required, required values can't be: None, '', [], {}.

    Raises:
        RequiredError: When value is required and match with False conditions.
    """
    if value in (Undefined, None, '', (), [], {}):
        raise RequiredError(f'The {attr_name} attribute is required.')


def _validate(attr_name: str, value: Any, attr_type: type) -> None: # pylint: disable=too-many-branches, too-many-return-statements
    """
    Validates that the given value is of the expected attribute type. The function supports special type checks for
    Date and DateTime types and handles general type validation for other types. It allows the value to pass through
    if it matches the expected type, is None, or is an instance of Undefined. The function uses custom checks for Date
    and DateTime to accommodate their unique validation requirements.

    Args:
        attr_name (str):
            The name of the attribute being validated. This is used for generating error messages.

        value (Any):
            The value to be validated against the expected type.

        attr_type (type):
            The expected type of the value. This can be a standard Python type, a custom type, or
            specific types like Date and DateTime which have dedicated validation logic.

    Raises:
        FieldValueError: When value is not of value_type.
    """
    # We always accept these 2 values
    if value is None or value is Undefined:
        return None

    if attr_type == Date:
        # If we are expecting a Date we don't need to check other things
        if Date.is_date(value):
            return None

    if attr_type == DateTime:
        # If we are expecting a DateTime we don't need to check other things
        if DateTime.is_datetime(value):
            return None

    # TypeError: typing.Any cannot be used with isinstance()
    if attr_type is Any:
        return None

    # https://everysk.atlassian.net/browse/COD-4286
    # If we use 'class'/callable as a annotation, the isinstance will fail
    # because attr_type will be a string/function so we need to discard it first
    if attr_type is callable:
        if callable(value):
            return None

    if isinstance(attr_type, str):
        if type(value).__name__ == attr_type:
            return None

    try:
        # When the second attribute is not a type or a tuple of types
        # the isinstance will fail with TypeError
        if isinstance(value, attr_type):
            return None
    except TypeError:
        pass

    raise FieldValueError(f'Key {attr_name} must be {attr_type}.')


class BaseField:
    """ Base class of all fields that will guarantee their type. """
    ## Public attributes
    attr_type: type = None
    default: Any = None
    readonly: bool = False
    required: bool = False
    required_lazy: bool = False
    empty_is_none: bool = False

    def __init__(
        self,
        attr_type: type,
        default: Any = None,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        """
        Use kwargs to set more attributes on the Field.

        Raises:
            DefaultError: For default values they can't be empty [] or {}.
            RequiredError: If field is readonly, then default value is required.
        """
        self.attr_type = attr_type
        if required and required_lazy:
            raise FieldValueError("Required and required_lazy can't be booth True.")

        self.required = required
        self.required_lazy = required_lazy

        if default is not None and not default and isinstance(default, (list, dict)):
            # For default values they can't be empty [] or {} - because this can cause
            # some issues with class attributes where these type can aggregate values.
            raise DefaultError('Default value cannot be a list or a dict.')

        if readonly and (default is None or default is Undefined):
            raise RequiredError('If field is readonly, then default value is required.')

        self.readonly = readonly

        # We use this flag to convert '' to None
        self.empty_is_none = empty_is_none

        # Other attributes will be assigned directly
        for key, value in kwargs.items():
            setattr(self, key, value)

        # For the last We need to store the cleaned value
        self.default = self.get_cleaned_value(default)

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __eq__(self, obj: object) -> bool:
        """
        One field will be equal to another one if all attributes are the same.

        Args:
            obj (object): The obj for compare.
        """
        return self.__dict__ == obj.__dict__

    def transform_to_none(self, value: Any) -> Any:
        """
        Transforms value to None if needed.
        """
        if self.empty_is_none and value == '':
            value = None

        return value

    def get_cleaned_value(self, value: Any) -> Any:
        # We first verify if we need to transform the value to None
        value = self.transform_to_none(value)

        # Then we run the get_value method when value is a callable
        value = self.get_value(value)

        # Then we run the clean_value method
        value = self.clean_value(value)

        return value

    def get_value(self, value: Any) -> Any:
        """
        Must be implemented in child classes that need do some changes
        on received value before clean_value.
        """
        # If value is callable we call it
        if callable(value):
            value = value()
        return value

    def clean_value(self, value: Any) -> Any:
        """
        Must be implemented in child classes that need do some
        changes on received value.
        """
        return value

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is required and if is of correct type.

        Raises:
            RequiredError: If required and None is passed
            FieldValueError: If value type don't match with required type.
        """
        if attr_type is None:
            attr_type = self.attr_type

        if self.readonly:
            # This is necessary to be able to at least assign the default value to the field
            if value != self.default:
                raise FieldValueError(f"The field '{attr_name}' value cannot be changed.")

        if self.required and not self.required_lazy:
            _required(attr_name, value)

        _validate(attr_name, value, attr_type)

    def __getattr__(self, name: str) -> Any:
        """
        This method is used to handle pylint errors where the method/attribute does not exist.
        The problem is that we change the field in the MetaClass to the Field's default value,
        so StrField does not have the str methods but the result is a string.
        This method will only be executed if the method/attributes do not exist in the Field class.

        Args:
            name (str): The name of the method/attribute.
        """
        # https://pythonhint.com/post/2118347356810295/avoid-pylint-warning-e1101-instance-of-has-no-member-for-class-with-dynamic-attributes
        return getattr(self.attr_type, name)


class MetaClass(type):
    _attr_name: str = '__attributes__'
    _anno_name: str = '__annotations__'

    def __call__(cls, *args: tuple, **kwargs: dict) -> Self:
        """
        Method that creates a sequence of class call like:
            before_init
            init
            after_init
        If the BaseObject class implement one of these methods they will be executed.
        https://discuss.python.org/t/add-a-post-method-equivalent-to-the-new-method-but-called-after-init/5449/11
        """
        errors: dict[str: Exception | None] = {'before_init': None, 'init': None, 'after_init': None}
        # We get the silent from kwargs otherwise we get from the class
        silent = kwargs.pop('silent', cls._silent)

        ## If the before init method is implemented we run it
        try:
            # If the method returns a dict we update kwargs
            dct = cls.__before_init__(**kwargs)
            if dct and isinstance(dct, dict):
                kwargs.update(dct)
        except Exception as error: # pylint: disable=broad-exception-caught
            if not silent:
                raise error
            errors['before_init'] = deepcopy(error)

        ## Here we create the object and initialize
        obj = cls.__new__(cls, *args, **kwargs)
        try:
            obj.__init__(*args, **kwargs)
        except Exception as error: # pylint: disable=broad-exception-caught
            if not silent:
                raise error
            errors['init'] = deepcopy(error)

        ## If the after init method is implemented we run it
        try:
            obj.__after_init__()
        except Exception as error: # pylint: disable=broad-exception-caught
            if not silent:
                raise error
            errors['after_init'] = deepcopy(error)

        # Store the errors
        if any(errors.values()):
            obj._errors = errors # pylint: disable=attribute-defined-outside-init
            # Execute the handler for errors
            obj._init_error_handler(kwargs, errors)

        # Set the instance to be Frozen or not
        try:
            obj._is_frozen = obj.config.frozen # pylint: disable=attribute-defined-outside-init
        except AttributeError:
            pass

        return obj

    def __new__(mcs, name: str, bases: tuple, attrs: dict) -> Self:
        """
        This method is executed every time a BaseObject Class is created in the Python runtime.
        We changed this method to create the config and attributes properties and update the annotations.

        Example:

            >>> from everysk.core.object import BaseObject
            >>> class MyClass(BaseDict):
            ...     class Config:
            ...         value: int = 1

            >>> obj1 = MyClass()
            >>> obj2 = MyClass()
            >>> MyClass.config.value, obj1.config.value, obj2.config.value
            (1, 1, 1)

            >>> obj1.config.value = 3
            >>> MyClass.config.value, obj1.config.value, obj2.config.value
            (3, 3, 3)

            >>> MyClass.Config
            ---------------------------------------------------------------------------
            AttributeError                            Traceback (most recent call last)
            ----> 1 MyClass.Config

            AttributeError: type object 'MyClass' has no attribute 'Config'

        Args:
            mcs (type): Represents this class.
            name (str): The name for the new class.
            bases (tuple): All inheritance classes.
            attrs (dict): All attributes that the new class has.

        Returns:
            type: Return the new class object.
        """
        # Creating the config property.
        try:
            Config = attrs.pop('Config') # pylint: disable=invalid-name
            attrs['config'] = Config()
        except KeyError:
            pass

        # We need all attributes to validate the types later
        # So we need to get the parents attributes too
        attributes = {}
        for parent in bases:
            attributes.update(getattr(parent, mcs._attr_name, {}))

        # We could not update the info inside the original attrs dict because:
        # RuntimeError: dictionary changed size during iteration
        # So we remove the attributes that we need to update
        attributes.update(attrs.pop(mcs._attr_name, {}))
        annotations: dict = attrs.pop(mcs._anno_name, {})
        new_attrs = {}
        for attr_name, attr_value in attrs.items():
            # We discard all python dunder attributes, all functions, all properties, Undefined and None values
            if not attr_name.startswith('__') and \
               not isroutine(attr_value) and \
               not isinstance(attr_value, property) and \
               attr_value is not None and \
               attr_value is not Undefined:
                # For BaseFields we need to use the properties and set the correct value in the attribute
                if isinstance(attr_value, BaseField):
                    # We keep a copy of the value inside the __attributes__
                    attributes[attr_name] = attr_value

                    # We set the correct value to this attribute in the class
                    new_attrs[attr_name] = attr_value.default

                    # We create the annotation for this attribute
                    if attr_name not in annotations:
                        annotations[attr_name] = attr_value.attr_type
                else:
                    # For normal attributes we only store the class
                    attributes[attr_name] = type(attr_value)

                    # Now we update annotations for attributes that are not annotated
                    # Ex: var = 1
                    if attr_name not in annotations:
                        annotations[attr_name] = type(attr_value)

        # With both completed now we need to get the fields that are only annotations
        # class MyClass:
        #     var: str
        for key in annotations.keys() - attributes.keys():
            attributes[key] = annotations[key]
            # We set the default value to None to avoid break the code
            new_attrs[key] = None

        # Then we update the attributes list for this new class
        attrs[mcs._attr_name] = attributes
        attrs[mcs._anno_name] = annotations
        attrs.update(new_attrs)

        return super().__new__(mcs, name, bases, attrs)

    def __setattr__(cls, __name: str, __value: Any) -> None:
        """
        Method that sets the values on fields of the class.

        Args:
            __name (str): The attribute name.
            __value (Any): The value that is set.
        """
        return super().__setattr__(__name, __get_field_value__(cls, __name, __value))


class BaseObject(metaclass=MetaClass):
    """
    To ensure correct check for data keys
    uses https://docs.python.org/3/library/typing.html standards.

        >>> from utils.object import BaseObject
        >>> class MyObject(BaseObject):
        ...     var_int: int = None
        ...
        >>> obj = MyObject(var_int='a')
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "/var/app/utils/object.py", line 151, in __init__
            setattr(self, key, value)
        File "/var/app/utils/object.py", line 209, in __setattr__
            value = self.__get_clean_value__(__name, value)
        File "/var/app/utils/object.py", line 192, in __get_clean_value__
            _validate(attr, value, annotations[attr])
        File "/var/app/utils/object.py", line 56, in _validate
            raise DataTypeError(f'Key {attr_name} must be {attr_type}.')
        utils.exceptions.DataTypeError: Key var_int must be <class 'int'>.
    """
    _errors: dict = None # Keep the init errors.
    _is_frozen: bool = False # This will control if we can update data on this class.
    _silent: bool = False # If true an error that happen on init will be stored in self._error.

    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls)

    @classmethod
    def __before_init__(cls, **kwargs: dict) -> dict:
        """
        Method that runs before the __init__ method.
        If needed we must return the kwargs that will be passed to the init
        otherwise return None to not change the current behavior.
        """
        return kwargs

    def __init__(self, **kwargs: dict) -> None:
        # Validate all required fields
        attributes = self.__get_attributes__()
        for attr_name, field in attributes.items():
            if getattr(field, 'required', False):
                _required(attr_name=attr_name, value=kwargs.get(attr_name))

        # Set all kwargs on the object
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __after_init__(self) -> None:
        """
        Method that runs after the __init__ method.
        This method must return None.
        """
        pass

    def __check_frozen__(self) -> None:
        """
        Method that checks if this class is a Frozen object and raises attribute error.
        """
        if self._is_frozen:
            raise AttributeError(f'Class {self.get_full_doted_class_path()} is frozen and cannot be modified.')

    def __copy__(self) -> Self:
        """
        A shallow copy constructs a new compound object and then (to the extent possible)
        inserts references into it to the objects found in the original.
        If the object is Frozen the copy will not be.
        This method is used when we call copy(obj).
        """
        # We need to copy the __dict__
        obj = copy(self.__dict__)

        # We must not copy the config attribute
        obj.pop('config', None)

        # We create a new obj
        obj = type(self)(**obj)
        return obj

    def __deepcopy__(self, memo: dict = None) -> Self:
        """
        A deep copy constructs a new compound object and then, recursively,
        inserts copies into it of the objects found in the original.
        This method is used when we call deepcopy(obj).

        Args:
            memo (dict, optional): A memory object to avoid copy twice. Defaults to None.
        """
        # We need to copy the __dict__
        obj = deepcopy(self.__dict__, memo)

        # We must not copy the config attribute
        obj.pop('config', None)

        # We create a new obj
        obj = type(self)(**obj)
        return obj

    def __delattr__(self, __name: str) -> None:
        """
        Method that removes __name from the object.

        Args:
            __name (str): The name of the attribute that will be removed.

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()
        super().__delattr__(__name)

    def __get_attributes__(self) -> dict:
        """
        Get all attributes from this class.
        """
        return getattr(self, MetaClass._attr_name) # pylint: disable=protected-access

    def __get_clean_value__(self, attr: str, value: Any) -> Any:
        """
        Pass value to a clean function and checks if value match's the correct type.

        Raises:
            FieldValueError: If value and type don't match the correct type.
        """
        return __get_field_value__(self, attr, value)

    def __getstate__(self) -> dict:
        """
        This method is used by Pickle module to get the correct serialized data.
        We need to remove the config attribute from the serialized data
        so we can recreate the object later.
        https://docs.python.org/3.11/library/pickle.html#handling-stateful-objects
        """
        dct = self.__dict__.copy()
        try:
            del dct['config']
        except KeyError:
            pass

        return dct

    def _init_error_handler(self, kwargs: dict, errors: dict[str, Exception]) -> None:
        """
        This method is called at the end of the init process if the silent param is True.

        Args:
            errors (dict): A dict {'before_init': None | Exception, 'init': None | Exception, 'after_init': None | Exception}
        """
        pass

    def __setattr__(self, __name: str, __value: Any) -> None:
        """
        Method changed from BaseClass for check de integrity of data.
        This method is executed on setting attributes in the object.
        Ex: obj.attr = 1

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()
        super().__setattr__(__name, self.__get_clean_value__(__name, __value))

    @classmethod
    def __set_attribute__(cls, attr_name: str, attr_type: Any, attr_value: Any) -> None:
        """
        Method that updates the list of attributes/annotations for this class.
        Normally this is used to update a class after it was created.

        Args:
            attr_name (str): The name for the new attribute.
            attr_type (Any): The type for the new attribute.
            attr_value (Any): The value for the new attribute.
        """
        attributes = getattr(cls, MetaClass._attr_name, {}) # pylint: disable=protected-access
        annotations = getattr(cls, MetaClass._anno_name, {}) # pylint: disable=protected-access
        try:
            # For BaseFields
            annotations[attr_name] = attr_type.attr_type
        except AttributeError:
            # Normal types
            annotations[attr_name] = attr_type

        attributes[attr_name] = attr_type
        # After we set the attributes/annotations we set the value in the class
        setattr(cls, attr_name, attr_value)

    ## Public methods
    def get_full_doted_class_path(self) -> str:
        """
        Return full doted class path to be used on import functions.
        Example:
            'everysk.core.BaseObject'
        """
        return f'{self.__module__}.{self.__class__.__name__}'

    def replace(self, **changes) -> Self:
        """
        Creates a new object of the same type as self, replacing fields with values from changes.

        Example:

            >>> from everysk.core.object import BaseObject
            >>> obj = BaseObject(attr=1)
            >>> obj.attr
            1
            >>> copy = obj.replace(attr=2)
            >>> copy.attr
            2

        Args:
            **changes (dict): All named params that are passed to this method.

        Returns:
            Self: A copy of this object with the new values.
        """
        obj = deepcopy(self.__dict__)

        # We must not copy the config attribute
        obj.pop('config', None)

        obj.update(changes)
        return type(self)(**obj)

    def validate_required_fields(self) -> None:
        """
        Try to validate all fields that are checked with required_lazy, because fields with
        required are always validate on the init.
        """
        attributes = self.__get_attributes__()
        for attr_name, field in attributes.items():
            if getattr(field, 'required_lazy', False):
                _required(attr_name=attr_name, value=getattr(self, attr_name, None))


class BaseDict(BaseObject, dict):
    """
    Extends BaseObject and also guarantees that BaseDict['key'] is equal to BaseDict.key

        >>> from utils.object import BaseDict
        >>> class MyDict(BaseDict):
        ...     var_int: int = None
        ...
        >>> obj = MyDict(var_int='test')
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "/var/app/utils/object.py", line 151, in __init__
            raise DataTypeError(f'Key {attr_name} must be {attr_type}.')
        File "/var/app/utils/object.py", line 279, in __setattr__
            if key.startswith('_'):
        File "/var/app/utils/object.py", line 209, in __setattr__
            DataTypeError: If value and type don't match the correct type.
        File "/var/app/utils/object.py", line 192, in __get_clean_value__
            # Create base annotations
        File "/var/app/utils/object.py", line 56, in _validate
            elif isinstance(value, attr_type):
        utils.exceptions.DataTypeError: Key var_int must be <class 'int'>.
        >>> obj = MyDict(var_int=10)
        >>> obj['var_int'] == obj.var_int
        True
    """
    _keys_blacklist: frozenset[str] = frozenset(['config'])

    ## Private Methods
    def __init__(self, **kwargs) -> None:
        ## After init we need to get all key/values from parents and set to the instance
        # Copy generates infinite recursion
        attributes = self.__get_attributes__()
        for key in attributes.keys() - kwargs.keys():
            if self.is_valid_key(key=key):
                # On init, attributes has value because inheritance
                # we are just updating the values of the keys
                # self['key'] is None but self.key has value
                self[key] = self[key]

        super().__init__(**kwargs)

    def __delattr__(self, name: str, caller: str = None) -> None:
        """
        Removes an atribute from self.

        Args:
            name (str): The attribute that will be removed.
            caller (str, optional): Used to avoid recursion when internally called. Defaults to None.

        Raises:
            AttributeError: If the attribute is not found.
        """
        # Remove attribute
        super().__delattr__(name)

        # Caller not None means this method was called from __delitem__
        # then we do not call the method again avoiding infinite loop
        if caller is None:
            try:
                # Some times the attribute will not exists
                self.__delitem__(name, caller='__delattr__')
            except KeyError:
                pass

    def __delitem__(self, key: Any, caller: str = None) -> None:
        """
        Removes an key from self.

        Args:
            key (Any): The key that will be removed.
            caller (str, optional): Used to avoid recursion when internally called. Defaults to None.

        Raises:
            KeyError: If the key is not found.
        """
        # Remove key
        super().__delitem__(key)

        # Caller not None means this method was called from __delattr__
        # then we do not call the method again avoiding infinite loop
        if caller is None:
            self.__delattr__(key, caller='__delitem__')

    def __getitem__(self, key: str) -> Any:
        """
        Get key from self or search on a parent.

        Args:
            key (str): The key name to search.

        Raises:
            KeyError: If the key does not exist.
        """
        if self.is_valid_key(key=key):
            try:
                return super().__getitem__(key)
            except KeyError:
                pass

            try:
                return getattr(self, key)
            except AttributeError:
                pass

        raise KeyError(key) # pylint: disable=raise-missing-from

    def __setattr__(self, name: str, value: Any, caller: str = None) -> None:
        """
        Method changed from BaseClass for guarantee de integrity of data attributes.
        This method is executed on setting attributes in the object.
        Ex: obj.attr = 1

        Raises:
            AttributeError: If is frozen.
        """
        super().__setattr__(name, value)
        # Because self.__get_clean_value__ can change the value we need to pick it from self
        new_value = getattr(self, name)

        # When value is associated:
        # directly on the key Ex dict['key'] = value then caller will be __setitem__
        # directly on attribute Ex dict.key = value then caller will be None
        # Don't do that to private attributes
        if caller is None and self.is_valid_key(key=name):
            # For integrity guarantee writes the value to the dictionary key as well.
            self.__setitem__(name, new_value, caller='__setattr__')

    def __setitem__(self, key: str, item: Any, caller: str = None) -> None:
        """
        Method changed from BaseClass for guarantee de integrity of data keys.
        This method is executed on setting items in the dictionary.
        Ex: d = dict(key=1) or d['key'] = 1

        Raises:
            AttributeError: If is frozen.
        """
        if key.startswith('_'):
            raise KeyError("Keys can't start with '_'")

        if key in self._keys_blacklist:
            raise KeyError(f'The key cannot be called "{key}".')

        # When value is associated:
        # directly on the key Ex dict['key'] = value then caller will be None
        # directly on attribute Ex dict.key = value then caller will be __setattr__
        if caller is None:
            # For integrity guarantee writes the value to the attribute as well.
            self.__setattr__(key, item, caller='__setitem__')

            # The setattr can "clean" the value them we need to catch it again
            item = getattr(self, key)

        # If key is a property we do not set on dict
        if not isinstance(getattr(self.__class__, key, None), property):
            super().__setitem__(key, item)

    def is_valid_key(self, key: str) -> bool:
        """
        This method checks if the key is valid.
        Valid keys are the ones that not starts with '_' and
        are not in the self._keys_blacklist.

        Args:
            key (str): The key that needs to be checked.
        """
        return not key.startswith('_') and key not in self._keys_blacklist

    ## Public Methods
    def clear(self) -> None:
        """
        This method clears the dictionary.

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()
        # Because we integrate key/attributes we need to remove booth
        # The original clear only remove keys
        # We need to convert keys to a list, because the original is a iterator
        for key in list(self.keys()):
            del self[key]

    def copy(self) -> dict:
        """
        Generate a copy for this object, we need to use deepcopy because
        this object could have some keys/attributes and they need to be on the copy.
        If the object is Frozen, the copy will not be.
        """
        return deepcopy(self)

    def fromkeys(self, keys: list, default: Any = None) -> dict:
        """
        Create a new object with the keys, if key does not exists add one with default value.
        If the object is Frozen, the copy will not be.

        Args:
            keys (list): The list with the keys for the new object.
            default (Any, optional): The default value if the key does not exists. Defaults to None.
        """
        # Create a new key list with all keys and use a set to avoid duplicates
        keys_aux = set(keys)
        keys_aux.update(self.keys())

        dct = self.copy()
        for key in keys_aux:
            if key in keys:
                dct[key] = dct.get(key, default)
            else:
                del dct[key]

        return dct

    def pop(self, *args) -> Any:
        """
        Remove specified key and return the corresponding value.
        If the key is not found return the default otherwise raise a KeyError.

        Args:
            key (str): The key that will be removed.
            default (Any, optional): The default value if key is not found.

        Raises:
            AttributeError: If is frozen.
            KeyError: If the default is not passed and key is not found.
        """
        self.__check_frozen__()
        ret = super().pop(*args)

        try:
            # Some times the attribute will not exists
            self.__delattr__(args[0]) # pylint: disable=unnecessary-dunder-call
        except AttributeError:
            pass

        return ret

    def popitem(self) -> tuple:
        """
        Remove and return a (key, value) pair as a 2-tuple.
        Pairs are returned in LIFO (last-in, first-out) order.

        Raises:
            AttributeError: If is frozen.
            KeyError: If the dict is empty.
        """
        self.__check_frozen__()
        return super().popitem()

    def update(self, *args, **kwargs) -> None:
        """
        Refactor to use __setitem__ and perform correct checks on data.
        https://stackoverflow.com/a/2588648
        https://stackoverflow.com/a/2390889

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()
        for key, value in dict(*args, **kwargs).items():
            self[key] = value



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import unittest
from everysk.config import settings

# from everysk.core.datetime import DateTime, ZoneInfo

from everysk.sdk.entities.worker_template.base import WorkerTemplate

from everysk.sdk.base import _mount_self_obj

###############################################################################
#   Report TestCase Implementation
###############################################################################
class WorkerTemplateTestCase(unittest.TestCase):

    def setUp(self):
        self.sample_data = {
            'created_on': None,
            'updated_on': None,
            'version': 'v1',
            'id': 'wrkt_xyz',
            'sort_index': 1,
            'category': 'portfolio',
            'form_outputs': {
                'OUTPUT_ONE': [
                    {
                        'key': 'value',
                    }
                ]
            },
            'script_entry_point': 'main',
            'tags': [],
            'type': 'BASIC',
            'script_visible_source': "def main(args):\n    print('test')\n    return",
            'form_functions': {},
            'ports': [
                {
                    'key': 'value',

                }
            ],
            'name': 'Worker Teste',
            'script_source': "def main(args):\n    print('test')\n    return",
            'icon': 'retriever',
            'script_runtime': 'python',
            'description': '#Worker Sample',
            'form_inputs': [
                {
                    'key': 'value'
                }

            ],
            'visible': True,
            'default_output': 'OUTPUT_ONE',
        }
        self.worker_template = WorkerTemplate(**self.sample_data)

    def test_get_id_prefix(self):
        self.assertEqual(WorkerTemplate.get_id_prefix(), settings.WORKER_TEMPLATE_ID_PREFIX)

    def test_validate(self):
        worker_template: WorkerTemplate = self.worker_template.copy()
        with unittest.mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = True
            result = worker_template.validate()

        mock_http_connection.assert_called_once_with(class_name='WorkerTemplate', params={}, method_name='validate', self_obj=_mount_self_obj(worker_template))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertTrue(result)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock

import os
import requests

from everysk.sdk.entities.portfolio.base import Portfolio
from everysk.core.datetime import Date, DateTime
from everysk.core.serialize import dumps
from everysk.core.object import BaseDict
from everysk.core.exceptions import HttpError, SDKError
from everysk.core.http import HttpSDKPOSTConnection

from everysk.sdk.base import BaseSDK, _mount_self_obj, _parser_out

###############################################################################
#   BaseSDK TestCase Implementation
###############################################################################
class TestBaseSDK(TestCase):

    class CustomObjectWithToDict:
        def __init__(self, value):
            self.value = value

        def to_dict(self, with_internals=False):
            return {'value': self.value} if not with_internals else {'value': self.value, 'internal': 'data'}

    def setUp(self):
        self.default_kwargs = {
            'class_name': 'BaseSDK',
            'method_name': 'setUp',  # This will be overridden in each test method.
            'self_obj': None,
            'params': {}
        }
        self.old_post = requests.post
        self.headers = {
            'Accept-Encoding': 'gzip, deflate;q=0.9',
            'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
            'Cache-control': 'no-cache',
            'Connection': 'close',
            'Content-Type': 'application/x-www-form-urlencoded',
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
            'Authorization': 'Bearer SID:TOKEN'
        }
        os.environ['EVERYSK_API_URL_SCHEME'] = 'https'
        os.environ['EVERYSK_API_URL_DOMAIN'] = 'test.com'
        os.environ['EVERYSK_API_SID'] = 'my_SID'
        os.environ['EVERYSK_API_TOKEN'] = 'my_TOKEN'
        response = mock.MagicMock()
        response.content = dumps({'my_SID': 'teste'})
        response.status_code = 200
        requests.post = mock.create_autospec(requests.post)
        requests.post.return_value = response

    def tearDown(self) -> None:
        requests.post = self.old_post

    def test_import_from_sdk_path(self):
        from everysk.sdk import Portfolio as root_Portfolio #pylint: disable=import-outside-toplevel

        self.assertEqual(root_Portfolio, Portfolio)

        with self.assertRaisesRegex(ImportError, "cannot import name 'Module' from 'everysk.sdk'"):
            from everysk.sdk import Module #pylint: disable=import-outside-toplevel

    def test_defaults_are_set(self):
        # Using the context manager to mock HttpSDKPOSTConnection
        http = BaseSDK.get_response(**self.default_kwargs)

        requests.post.assert_called_once_with(
            url='https://test.com/v1/sdk_function',
            headers=self.headers,
            verify=True,
            timeout=30,
            data=dumps(self.default_kwargs)
        )

    def test_automatic_class_name_assignment(self):
        # In this scenario, we are not passing 'class_name'.
        # Directly calling get_response from BaseSDK
        kwargs = {'method_name': 'test_automatic_class_name_assignment'}
        BaseSDK.get_response(**kwargs)

        expected_kwargs = {**self.default_kwargs, **kwargs, 'class_name': 'BaseSDK'}

        requests.post.assert_called_once_with(
            url='https://test.com/v1/sdk_function',
            headers=self.headers,
            verify=True,
            timeout=30,
            data=dumps(expected_kwargs)
        )

    def test_automatic_method_name_assignment(self):
        # In this scenario, we are not passing 'class_name'.
        # Directly calling get_response from BaseSDK
        kwargs = {'class_name': 'BaseSDK'}
        BaseSDK.get_response(**kwargs)

        expected_kwargs = {**self.default_kwargs, **kwargs, 'method_name': 'test_automatic_method_name_assignment'}

        requests.post.assert_called_once_with(
            url='https://test.com/v1/sdk_function',
            headers=self.headers,
            verify=True,
            timeout=30,
            data=dumps(expected_kwargs)
        )

    def test_get_response_raises_sdk_error_on_http_error(self):
        with mock.patch.object(HttpSDKPOSTConnection, 'get_response', side_effect=HttpError("Test HTTP Error")):
            with self.assertRaises(SDKError) as context:
                BaseSDK.get_response()
            self.assertEqual(str(context.exception), "Test HTTP Error")

    def test_get_response_with_sdk_error(self):
        with mock.patch.object(HttpSDKPOSTConnection, 'get_response', return_value={'error_module':'everysk.core.exceptions.SDKError', 'error_message':'SDK sample error'}):
            with self.assertRaisesRegex(SDKError, 'SDK sample error'):
                BaseSDK.get_response()

    def test_to_dict_returns_dict(self):
        class TestSDK(BaseSDK):
            def to_dict(self, with_internals: bool = True) -> dict:
                return super().to_dict(with_internals)

        test_sdk = TestSDK(
            test='test',
            date=Date(2021, 1, 1),
            datetime=DateTime(2021, 1, 1, 1, 1, 1),
            date_list=[Date(2021, 1, 1)],
            datetime_list=[DateTime(2021, 1, 1, 1, 1, 1)],
            dates_dict={'date': Date(2021, 1, 1), 'datetime': DateTime(2021, 1, 1, 1, 1, 1), 'date_list': [Date(2021, 1, 1)], 'datetime_list': [DateTime(2021, 1, 1, 1, 1, 1)]},
            tuple_list=[(Date(2021, 1, 1),)],
            date_tuple=(Date(2021, 1, 1),),
            date_set=set([Date(2021, 1, 1)]),
        )
        test_sdk.obj_list = [TestSDK(
            test='test',
            date=Date(2021, 1, 1),
            datetime=DateTime(2021, 1, 1, 1, 1, 1),
        )]
        self.assertDictEqual(test_sdk.to_dict(), {
            'test': 'test',
            'date': '20210101',
            'datetime': '20210101 01:01:01',
            'date_list': ['20210101'],
            'datetime_list': ['20210101 01:01:01'],
            'dates_dict': {'date': '20210101', 'datetime': '20210101 01:01:01', 'date_list': ['20210101'], 'datetime_list': ['20210101 01:01:01']},
            'tuple_list': [('20210101',)],
            'date_tuple': ('20210101',),
            'date_set': ['20210101'],
            'obj_list': [{
                'test': 'test',
                'date': '20210101',
                'datetime': '20210101 01:01:01',
            }]
        })

    def test_mount_self_obj_none(self):
        self.assertIsNone(_mount_self_obj(None))

    def test_mount_self_obj_with_private_attrs(self):
        class TestObject(BaseDict):
            def __init__(self):
                self._private = 1
                self.public = 2

        test_obj = TestObject()
        result = _mount_self_obj(test_obj)
        self.assertIn('_private', result)
        self.assertIn('public', result)
        self.assertEqual(result['_private'], 1)

    def test_mount_self_obj_with_serializable_object(self):
        class Serializable(BaseDict):
            def __init__(self):
                self.teste = 1
                self._klass = DateTime

        class TestObject(BaseDict):
            def __init__(self):
                self._private = Serializable()

        test_obj = TestObject()
        result = _mount_self_obj(test_obj)
        self.assertDictEqual(result, {'_private': {'teste': 1, '_klass': 'DateTime'}})

    def test_parser_out_with_primitive(self):
        self.assertEqual(_parser_out(42), 42)
        self.assertEqual(_parser_out("test"), "test")

    def test_parser_out_with_list(self):
        self.assertEqual(_parser_out([1, "a", [3, 4]]), [1, "a", [3, 4]])

    def test_parser_out_with_tuple(self):
        self.assertEqual(_parser_out((1, "b", (3, 4))), (1, "b", (3, 4)))

    def test_parser_out_with_set(self):
        self.assertEqual(_parser_out({1, 2, 3}), [1, 2, 3])  # Order might need to be considered

    def test_parser_out_with_dict(self):
        input_dict = {"a": 1, "b": {"c": 2}}
        self.assertDictEqual(_parser_out(input_dict), input_dict)

    def test_parser_out_custom_object_with_to_dict(self):
        obj = self.CustomObjectWithToDict('data')
        self.assertDictEqual(_parser_out(obj, False), {'value': 'data'})
        self.assertDictEqual(_parser_out(obj), {'value': 'data', 'internal': 'data'})

    def test_parser_out_custom_object_with_to_dict_in_list(self):
        obj_list = [self.CustomObjectWithToDict('data1'), self.CustomObjectWithToDict('data2')]
        expected_output = [{'value': 'data1'}, {'value': 'data2'}]
        self.assertListEqual(_parser_out(obj_list, False), expected_output)

    def test_parser_out_custom_object_with_strftime_or_null(self):
        self.assertEqual(_parser_out(DateTime(2020, 1, 1)), '20200101 00:00:00')

    def test_parser_out_with_undefined(self):
        self.assertEqual(_parser_out(Undefined), Undefined.default_parse_string)

    def test_parser_out_with_internals_flag(self):
        obj = self.CustomObjectWithToDict('data')
        self.assertDictEqual(_parser_out(obj, with_internals=False), {'value': 'data'})
        self.assertDictEqual(_parser_out(obj, with_internals=True), {'value': 'data', 'internal': 'data'})



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.fields import StrField, ChoiceField, ListField, BoolField, DictField, IntField

from everysk.sdk.entities.base import BaseEntity
from everysk.sdk.entities.fields import EntityNameField, EntityTagsField

###############################################################################
#   Worker Template Implementation
###############################################################################
class WorkerTemplate(BaseEntity):

    description = StrField()
    tags = EntityTagsField()
    name = EntityNameField()

    visible = BoolField(default=False)
    category = ChoiceField(default=Undefined, choices=settings.WORKER_TEMPLATE_CATEGORIES, required_lazy=True)
    icon = ChoiceField(default=Undefined, choices=settings.WORKER_TEMPLATE_ICONS, required_lazy=True)
    type = ChoiceField(default=settings.WORKER_TEMPLATE_DEFAULT_TYPE, choices=settings.WORKER_TEMPLATE_TYPES, required_lazy=True)
    sort_index = IntField(required_lazy=True)

    ports = ListField()
    form_functions = DictField()
    form_inputs = ListField(max_size=settings.WORKER_TEMPLATE_SCRIPT_OUTPUTS_MAX_SIZE)
    form_outputs = DictField()
    default_output = StrField(required_lazy=True)

    script_runtime = ChoiceField(default=settings.WORKER_TEMPLATE_DEFAULT_SCRIPT_RUNTIME, choices=settings.WORKER_TEMPLATE_SCRIPT_RUNTIMES, required_lazy=True)
    script_entry_point = StrField(required_lazy=True)

    script_source = StrField(min_size=settings.WORKER_TEMPLATE_SCRIPT_SOURCE_MIN_SIZE, max_size=settings.WORKER_TEMPLATE_SCRIPT_SOURCE_MAX_SIZE, required_lazy=True)
    script_visible_source = StrField(min_size=settings.WORKER_TEMPLATE_SCRIPT_SOURCE_MIN_SIZE, max_size=settings.WORKER_TEMPLATE_SCRIPT_SOURCE_MAX_SIZE)

    @staticmethod
    def get_id_prefix() -> str:
        """
        Get the prefix for the unique identifier for this entity.

        Returns:
            str: The prefix for the unique identifier.

        Raises:
            NotImplementedError: This method should be overridden in subclasses.
        """
        return settings.WORKER_TEMPLATE_ID_PREFIX

    def validate(self) -> bool:
        """
        Validate the entity properties.

        Raises:
            FieldValueError: If the entity properties are invalid.
            RequiredFieldError: If a required field is missing.

        Returns:
            bool: True if the entity properties are valid, False otherwise.

        Example usage:
            >>> entity = MyEntity()
            >>> entity.validate()
            Traceback (most recent call last):
                ...
            everysk.sdk.exceptions.RequiredFieldError: The field 'name' is required.
        """
        return self.get_response(self_obj=self)



# -*- coding: utf_8 -*-
###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Any, TYPE_CHECKING

import csv
import six

from everysk.config import settings

from everysk.core.string import to_string
from everysk.core.fields import ListField, StrField, FloatField, BoolField

from everysk.sdk.entities.base import BaseEntity, ScriptMetaClass
from everysk.sdk.entities.script import Script
from everysk.sdk.entities.fields import EntityNameField, EntityDescriptionField, EntityLinkUIDField, EntityWorkspaceField, EntityDateTimeField, EntityTagsField, CurrencyField
from everysk.sdk.entities.portfolio.securities import Securities

if TYPE_CHECKING:
    from io import StringIO

###############################################################################
#   Securities Field Implementation
###############################################################################
class SecuritiesField(ListField):
    attr_type = Securities

    def __init__(self, default: Any = None, required_lazy: bool = True, **kwargs) -> None:
        """
        This method initializes the field.

        Args:
            default (Any): The default value of the field.
            required_lazy (bool): True if the field is required, False otherwise.
            **kwargs (dict): The keyword arguments of the field.
        """
        super().__init__(default=default, required_lazy=required_lazy, **kwargs)

    def clean_value(self, value: Any) -> Securities:
        """
        This method cleans the value of the field.

        Args:s
            value (Any): The value to be cleaned.

        Returns:
            Securities: The cleaned value.
        """
        if isinstance(value, list):
            value = self.attr_type(value)
        return super().clean_value(value)

###############################################################################
#   Portfolio Implementation
###############################################################################
class Portfolio(BaseEntity, metaclass=ScriptMetaClass):
    """
    This class represents a portfolio entity object and provides methods to validate
    and convert the entity object to a dictionary or a CSV string representation.

    Attributes:
        script (Script): The script object associated to the entity.
        id (StrField): The unique identifier of the portfolio.
        workspace (EntityWorkspaceField): The workspace of the portfolio.
        name (EntityNameField): The name of the portfolio.
        tags (EntityTagsField): The tags of the portfolio.
        link_uid (EntityLinkUIDField): The link uid of the portfolio.
        description (EntityDescriptionField): The description of the portfolio.
        nlv (FloatField): The nlv of the portfolio.
        base_currency (CurrencyField): The base currency of the portfolio.
        date (EntityDateTimeField): The date of the portfolio.
        securities (SecuritiesField): The securities of the portfolio.
        level (StrField): The level of the portfolio.

        version (StrField): The version of the portfolio.
        created_on (DateTimeField): The created on date of the portfolio.
        updated_on (DateTimeField): The updated on date of the portfolio.

    Example usage:
        >>> from everysk.sdk.entities.portfolio.base import Portfolio
        >>> portfolio = Portfolio()
        >>> portfolio.workspace = 'my_workspace'
        >>> portfolio.name = 'My Portfolio'
        >>> portfolio.tags = ['tag1', 'tag2']
        >>> portfolio.nlv = 1000.0
        >>> portfolio.base_currency = 'USD'
        >>> portfolio.date = '20210101'
        >>> portfolio.securities = ['sec1', 'sec2', 'sec3']
        >>> portfolio.save()
        >>> portfolio.to_dict()
        {
            'id': 'port_12345678',
            'workspace': 'my_workspace',
            'name': 'My Portfolio',
            'tags': ['tag1', 'tag2'],
            'nlv': 1000.0,
            'base_currency': 'USD',
            'date': '20210101',
            'securities': ['sec1', 'sec2', 'sec3'],
            'version': 'v1',
            'created_on': '2021-01-01T00:00:00.000000Z',
            'updated_on': '2021-01-01T00:00:00.000000Z',
            'level': 'portfolio'
        }

    """
    script: Script
    _orderable_attributes = ListField(default=['date', 'created_on', 'updated_on', 'name']) # TODO changed to ReadonlyListField

    id = StrField(regex=settings.PORTFOLIO_ID_REGEX, max_size=settings.PORTFOLIO_ID_MAX_SIZE, required_lazy=True, empty_is_none=True)

    name = EntityNameField()
    description = EntityDescriptionField()
    tags = EntityTagsField()
    link_uid = EntityLinkUIDField()
    workspace = EntityWorkspaceField()

    nlv = FloatField(min_size=0.0, max_size=float('inf'))
    base_currency = CurrencyField()
    date = EntityDateTimeField()
    securities = SecuritiesField(min_size=0, max_size=settings.PORTFOLIO_MAX_SIZE, required_lazy=True)
    level = StrField(min_size=1, max_size=settings.PORTFOLIO_LEVEL_MAX_LENGTH)

    check_securities = BoolField(default=False)

    ###############################################################################
    #   Portfolio Basics Implementation
    ###############################################################################

    @staticmethod
    def get_id_prefix() -> str:
        """
        Returns the prefix of the portfolio id field value.

        Returns:
            str: The prefix of the portfolio id field value.

        Usage:
            >>> Portfolio.get_id_prefix()
            'port_'

        Notes:
            The prefix is typically used to distinguish portfolio IDs from other types of IDs.
        """
        return settings.PORTFOLIO_ID_PREFIX

    def validate(self) -> bool:
        """
        This method validates the entity object and raises an exception if it is not
        valid. The validation is performed by calling the `validate` method of each field
        of the entity.

        Args:
            self (Self): The entity object to be validated.

        Raises:
            FieldValueError: If the entity object is not valid.
            RequiredFieldError: If a required field is missing.

        Returns:
            bool: True if the entity object is valid, False otherwise.

        Example usage:
            >>> entity = Entity()
            >>> entity.validate()
            True
        """
        return self.get_response(self_obj=self)

    def to_csv(self) -> str:
        """
        This method converts the entity object into a CSV string representation of the
        entity object.

        Returns:
            str: A CSV string representation of the entity.

        Example usage:
            >>> entity = Entity()
            >>> entity.to_csv()
            "name,date,base_currency,nlv,description,tags,securities
            My Portfolio,2021-01-01,USD,1000.0,My Portfolio Description,tag1 tag2,sec1,sec2,sec3"
        """
        def write_row(csv_writer_: Any, line: list[str], length: int) -> None:
            """
            Write a row to a CSV file

            Args:
                csv_writer_ (Any): CSV writer object to use for writing the row
                line (list[str]): The list of values to write as a row
                length (int): The desired length of the row

            Returns:
                None

            Usage:
                >>> from everysk.sdk.entities.portfolio.base import Portfolio
                >>> import csv
                >>> csv_file = open('output'.csv', 'w', newline='')
                >>> write = csv.writer(csv_file)
                >>> Portfolio.write_row(writer, ['value1', 'value2'], 5)
                >>> csv_file.close()

            Notes:
                This function pads the row with empty strings to ensure the fixed length.
            """
            out: list[str] = line + ([''] * (length - len(line)))
            out = [to_string(s) if s is not None else None for s in out]
            csv_writer_.writerow(out)
            return None

        csv_file: StringIO = six.StringIO()
        csv_writer: Any = csv.writer(csv_file)

        date_: str = self.date.strftime()
        nlv_: str = '' if not self.nlv else to_string(self.nlv)
        pre_headers: list[str] = [self.name, date_, self.base_currency, nlv_, self.description, ' '.join(self.tags)]

        securities_as_lists: list[list[str]] = self.securities.to_lists()

        headers: list[str] = securities_as_lists[0]
        headers_len: int = len(headers)
        write_row(csv_writer, pre_headers, headers_len)
        write_row(csv_writer, headers, headers_len)

        for sec_as_list in securities_as_lists[1:]:
            write_row(csv_writer, sec_as_list, headers_len)

        out: str = csv_file.getvalue()
        csv_file.close()
        return out



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Any, Callable

from everysk.sdk.base import BaseSDK
###############################################################################
#   Base Script Implementation
###############################################################################
class Script(BaseSDK):
    """
    A base class for scripted queries.
    This class provides a base implementation for scripted queries.

    Attributes:
        - _klass (callable): The class to instantiate when fetching an entity.

    Example usage:
        To fetch an entity:
        >>> script = Script(klass=MyEntity)
        >>> entity = script.fetch(user_input, variant, workspace)
    """
    _klass: Callable = None

    def __init__(self, _klass: Callable) -> None:
        super().__init__(_klass=_klass)

    def fetch(self, user_input: Any, variant: str, workspace: str) -> Any:
        """
        Process a scripted query based on user input, variant, and workspace.

        This method provides a way to construct and execute different types of queries
        based on the specified variant. It's designed to handle a variety of scenarios
        and return the desired entity or entities based on the input parameters.

        Parameters:
            - user_input (Any): The input provided by the user, which can be used for filtering
            or as a direct entity ID.
            - variant (str): The type of scripted query to execute. Determines how the method
            processes the user input and constructs the query. Supported variants include
            'previousWorkers', 'tagLatest', any string starting with 'select', and potentially
            others.
            - workspace (str): The workspace context for the query. Used for scoping and
            verifying entity retrieval.

        Returns:
            - Any: Depending on the variant and user input, the method might return an entity,
            a list of entities, or None.

        Raises:
            - ValueError: If there's an attempted cross-workspace operation or other variant-specific
            error conditions are met.

        Note:
            The method behavior can vary greatly depending on the `variant` parameter, and it's
            important to ensure that the variant aligns with the expected user input structure.

        Example usage:
            To fetch an entity:
            >>> script = Script(klass=MyEntity)
            >>> entity = script.fetch(user_input, variant, workspace)
        """
        if user_input is None:
            return None

        if variant == 'previousWorkers' and user_input.get('id') is None:
            return self._klass(**user_input)

        return self.get_response(self_obj=self, params={'user_input': user_input, 'variant': variant, 'workspace': workspace})

    def persist(self, entity: Any, persist: str) -> Any:
        """
        This method provides a way to persist an entity based on the specified persist type.

        Parameters:
            - entity (Any): The entity to persist.
            - persist (str): The type of persist to execute. Determines how the method
            persists the entity. Supported persists include 'insert', 'update', and 'delete'.

        Returns:
            - Any: Depending on the persist type, the method might return an entity.

        Example usage:
            To persist an entity:
            >>> script = Script(klass=MyEntity)
            >>> entity = script.persist(entity, persist)
        """
        return self.get_response(self_obj=self, params={'entity': entity, 'persist': persist})



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase

from everysk.config import settings
from everysk.core.datetime import Date
from everysk.core.exceptions import FieldValueError

from everysk.sdk.entities.portfolio.security import Security

###############################################################################
#   Security TestCase Implementation
###############################################################################
class TestSecurity(TestCase):

    def setUp(self):
        # Sample data for valid initialization
        self.valid_data = {
            'status': 'OK',
            'symbol': 'AAPL',
            'id': 'unique123',
            'quantity': 100.0,
            'instrument_class': 'Equity',
            'maturity_date': Date(2025, 5, 1),
            'issue_date': Date(2020, 1, 1),
            'return_date': Date(2025, 12, 31),
            'settlement': Date(2020, 1, 2),
            'display': 'Apple Inc.',
            'comparable': 'Yes',
            'previous_quantity': 50.0,
            'extra_field1': 'extra_value1',  # This should be placed in `extra_data`
            'extra_field2': 'extra_value2'   # This should also be placed in `extra_data`
        }
        self.security = Security(**self.valid_data)

    def test_initialization_with_valid_data(self):
        security = Security(**self.valid_data)
        self.assertEqual(security.status, 'OK')
        self.assertEqual(security.id, 'unique123')
        self.assertEqual(security.extra_data['extra_field1'], 'extra_value1')
        self.assertEqual(security.extra_data['extra_field2'], 'extra_value2')

    def test_from_list_conversion(self):
        headers = list(self.valid_data.keys())
        values = list(self.valid_data.values())
        security = Security.from_list(values, headers)
        self.assertEqual(security.status, 'OK')
        self.assertEqual(security.id, 'unique123')
        self.assertEqual(security.extra_data['extra_field1'], 'extra_value1')

    def test_to_dict_conversion(self):
        security = Security(**self.valid_data)
        dict_data = security.to_dict()
        self.assertEqual(dict_data['status'], 'OK')
        self.assertEqual(dict_data['id'], 'unique123')
        self.assertEqual(dict_data['symbol'], 'AAPL')
        self.assertEqual(dict_data['quantity'], 100.0)
        self.assertEqual(dict_data['instrument_class'], 'Equity')
        self.assertDictEqual(dict_data['extra_data'], {'extra_field1': 'extra_value1', 'extra_field2': 'extra_value2'})
        # Add additional checks as required.

    def test_initialization_with_extra_data(self):
        data_with_extra = self.valid_data.copy()
        data_with_extra['random_key'] = 'random_value'
        security = Security(**data_with_extra)
        self.assertEqual(security.extra_data['random_key'], 'random_value')

    def test_symbol_length_validation(self):
        invalid_data = self.valid_data.copy()
        invalid_data['symbol'] = 'A' * (settings.SYMBOL_ID_MAX_LEN + 1)
        with self.assertRaisesRegex(FieldValueError, "The length '101' for attribute 'symbol' must be between '0' and '100'."):
            Security(**invalid_data)

    def test_security_id_length_validation(self):
        invalid_data = self.valid_data.copy()
        invalid_data['id'] = '1' * (settings.SYMBOL_ID_MAX_LEN + 1)
        with self.assertRaisesRegex(FieldValueError, "The length '101' for attribute 'id' must be between '1' and '100'."):
            Security(**invalid_data)

    def test_date_conversion_on_to_dict(self):
        valid_data_with_date = self.valid_data.copy()
        valid_date = Date.fromisoformat('2022-01-01')
        valid_data_with_date['maturity_date'] = valid_date
        security = Security(**valid_data_with_date)
        dict_data = security.to_dict()
        self.assertEqual(dict_data['maturity_date'], '20220101')

    def test_extra_data_initialization(self):
        # Test that fields not in the class annotations are moved to `extra_data`
        data_with_extra = self.valid_data.copy()
        data_with_extra['extra_data'] = {'random_field': 'random_value'}
        security = Security(**data_with_extra)
        self.assertIn('random_field', security.extra_data)
        self.assertEqual(security.extra_data['random_field'], 'random_value')

    def test_extra_data_empty_after_initialization(self):
        # Test that if no extra data is present, `extra_data` is set to None
        data_without_extra = self.valid_data.copy()
        data_without_extra.pop('extra_field1')
        data_without_extra.pop('extra_field2')
        security = Security(**data_without_extra)
        self.assertIsNone(security.extra_data)

    def test_generate_security_id(self):
        security_id1 = Security.generate_security_id()
        security_id2 = Security.generate_security_id()

        self.assertIsInstance(security_id1, str)
        self.assertIsInstance(security_id2, str)
        self.assertNotEqual(security_id1, security_id2)

    def test_sort_header(self):
        headers = ['status', 'symbol', 'quantity', 'instrument_class']
        headers_cmp = ['quantity', 'instrument_class', 'status', 'symbol']
        sorted_headers = Security.sort_header(headers_cmp)
        self.assertListEqual(sorted_headers, headers)

    def test_validate_required_fields_without_id(self):
        # Test if ID gets auto-generated if not provided
        data_without_id = self.valid_data.copy()
        data_without_id.pop('id')
        security = Security(**data_without_id)
        security.validate_required_fields()
        self.assertIsNotNone(security.id)

    def test_validate_required_fields_with_id(self):
        # Test if provided ID is not changed
        data_with_id = self.valid_data.copy()
        data_with_id['id'] = 'test_id'
        security = Security(**data_with_id)
        security.validate_required_fields()
        self.assertEqual(security.id, "test_id")

    def test_date_fields_to_string(self):
        json_representation = Security(**self.valid_data).to_dict()
        self.assertEqual(json_representation['maturity_date'], '20250501')
        self.assertEqual(json_representation['issue_date'], '20200101')
        self.assertEqual(json_representation['return_date'], '20251231')
        self.assertEqual(json_representation['settlement'], '20200102')

    def test_exclude_internal_fields(self):
        json_representation = Security(**self.valid_data).to_dict(with_internals=False)
        self.assertNotIn('display', json_representation)
        self.assertNotIn('comparable', json_representation)
        self.assertNotIn('previous_quantity', json_representation)

    def test_include_internal_fields(self):
        json_representation = Security(**self.valid_data).to_dict(with_internals=True)
        self.assertIn('display', json_representation)
        self.assertEqual(json_representation['display'], self.valid_data['display'])
        self.assertIn('comparable', json_representation)
        self.assertEqual(json_representation['comparable'], self.valid_data['comparable'])
        self.assertIn('previous_quantity', json_representation)
        self.assertEqual(json_representation['previous_quantity'], self.valid_data['previous_quantity'])

    def test_get_attr(self):
        ret = Security._get_attr(self.security, 'symbol') # pylint: disable=protected-access
        self.assertEqual(ret, 'AAPL')

    def test_get_attr_in_extra_data(self):
        valid_data = self.valid_data.copy()
        valid_data['extra_data'] = {'other_key': 'other_value'}
        security = Security(**valid_data)
        ret = Security._get_attr(security, 'other_key') # pylint: disable=protected-access
        self.assertEqual(ret, 'other_value')

    def test_get_attr_fallback(self):
        ret = Security._get_attr(self.security, 'inexistent_key', Security.generate_security_id) # pylint: disable=protected-access
        self.assertIsInstance(ret, str)

    def test_get_attr_inexistent(self):
        ret = Security._get_attr(self.security, 'inexistent_key') # pylint: disable=protected-access
        self.assertIsNone(ret)

    def test_generate_consolidation_key(self):
        ret = self.security.generate_consolidation_key(['symbol', 'instrument_class'])
        self.assertEqual(ret, 'AAPL_Equity')

    def test_to_list_without_headers(self):
        ret = self.security.to_list()
        self.assertListEqual(ret, ['OK', 'unique123', 'AAPL', 100.0, 'Equity', None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, '20250501', None, None, None, None, None, None, None, None, None, None, '20200101', None, None, None, None, None, None, None, None, None, None, None, '20251231', '20200102', None, {'extra_field1': 'extra_value1', 'extra_field2': 'extra_value2'}, None, 'Yes', 'Apple Inc.', None, None, 50.0])



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import unittest
from everysk.config import settings

from everysk.core.datetime import DateTime, ZoneInfo

from everysk.sdk.entities.report.base import Report

from everysk.sdk.base import _mount_self_obj

###############################################################################
#   Report TestCase Implementation
###############################################################################
class ReportTestCase(unittest.TestCase):

    def setUp(self):
        self.sample_data = {
            'created_on': None,
            'updated_on': None,
            'version': 'v1',
            'id': 'repo_12345678',
            'name': 'My Report',
            'description': 'This is a sample report.',
            'tags': ['tag1',
            'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': DateTime(2022, 1, 1, 12, 0, tzinfo=ZoneInfo(key='UTC')),
            'widgets': [{'type': 'chart', 'data': {}}],
            'url': '/sefdsf5s54sdfsksdfs5',
            'authorization': 'private',
            'config_cascaded': {'setting1': 'value1',
            'setting2': 'value2'},
            'layout_content': {'section1': {}}
        }
        self.report = Report(**self.sample_data)

    def test_get_id_prefix(self):
        self.assertEqual(Report.get_id_prefix(), settings.REPORT_ID_PREFIX)

    def test_validate(self):
        report: Report = self.report.copy()
        with unittest.mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = True
            result = report.validate()

        mock_http_connection.assert_called_once_with(class_name='Report', params={}, method_name='validate', self_obj=_mount_self_obj(report))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertTrue(result)

    def test_to_dict(self):
        dict_output = self.report.to_dict()
        self.assertDictEqual(dict_output, {
            'version': 'v1',
            'id': 'repo_12345678',
            'name': 'My Report',
            'description': 'This is a sample report.',
            'tags': ['tag1', 'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': '20220101',
            'widgets': [{'type': 'chart', 'data': {}}],
            'url': 'https://app.everysk.com/report/sefdsf5s54sdfsksdfs5',
            'authorization': 'PRIVATE',
            'date_time': '20220101 12:00:00',
            'created': None,
            'updated': None,
            'absolute_url': 'https://app.everysk.com/report/sefdsf5s54sdfsksdfs5',
            'relative_url': '/report/sefdsf5s54sdfsksdfs5'
        })

    def test_to_dict_no_modify_obj_by_reference(self):
        report = Report(**self.sample_data)
        dict_output = report.to_dict()

        dict_output['tags'].append('tag3')
        self.assertNotIn('tag3', report.tags)
        self.assertIn('config_cascaded', report)
        self.assertNotIn('config_cascaded', dict_output)
        self.assertIn('layout_content', report)
        self.assertNotIn('layout_content', dict_output)

        self.assertEqual('/sefdsf5s54sdfsksdfs5', report.url)
        self.assertEqual('private', report.authorization)

        self.assertDictEqual(report.config_cascaded, {'setting1': 'value1', 'setting2': 'value2'})
        self.assertDictEqual(report.layout_content, {'section1': {}})

    def test_query_load_with_id(self):
        with unittest.mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_data
            result = Report(id='repo_1234567891011211234567890', workspace='SampleWorkspace').load()

        mock_http_connection.assert_called_once_with(
            params={'entity_id': 'repo_1234567891011211234567890'},
            class_name='Report',
            method_name='retrieve',
            self_obj=None
        )
        mock_http_connection.return_value.get_response.assert_called_once_with()

        self.assertEqual(result, self.report)
        self.assertIsInstance(result, Report)

    def test_query_load(self):
        with unittest.mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_data
            result = Report(link_uid='SampleLinkUID', workspace='SampleWorkspace').load()

        mock_http_connection.assert_called_once_with(
            self_obj=_mount_self_obj({
                'filters': [('workspace', '=', 'SampleWorkspace'), ('link_uid', '=', 'SampleLinkUID')],
                'order': [],
                'projection': None,
                'limit': None,
                'offset': None,
                'page_size': None,
                'page_token': None,
                'distinct_on': [],
                '_klass': 'Report',
                '_clean_order': []
            }),
            params={'offset': None},
            class_name='Query',
            method_name='load'
        )
        mock_http_connection.return_value.get_response.assert_called_once_with()

        self.assertEqual(result, self.report)
        self.assertIsInstance(result, Report)

    def test_entity_to_query(self):
        cmp_query = Report.query.where('url', self.report.url)
        query = Report(url=self.report.url)._entity_to_query() #pylint: disable=protected-access

        self.assertDictEqual(query, cmp_query)



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import requests

from requests.packages.urllib3.exceptions import InsecureRequestWarning # pylint: disable=import-error
requests.packages.urllib3.disable_warnings(InsecureRequestWarning) # pylint: disable=no-member

###############################################################################
#   HTTPClient Implementation
###############################################################################
class HTTPClient(object):
    def __init__(self, timeout=3600, verify_ssl_certs=True, allow_redirects=False) -> None:
        self.timeout = timeout
        self.allow_redirects = allow_redirects
        self.verify_ssl_certs = verify_ssl_certs

    def request(self, method, url, headers, params=None, payload=None):
        raise NotImplementedError(
            'HTTPClient subclasses must implement `request`'
        )

###############################################################################
#   Requests Client Implementation
###############################################################################
class RequestsClient(HTTPClient):
    def request(self, method, url, headers, params=None, payload=None):

        response = requests.request(
            method,
            url,
            headers=headers,
            params=params,
            json=payload,
            timeout=self.timeout,
            verify=self.verify_ssl_certs,
            allow_redirects=self.allow_redirects
        )
        return (response.status_code, response.content)

def new_default_http_client(*args, **kwargs):
    return RequestsClient(*args, **kwargs)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock

from everysk.core.exceptions import RequiredError
from everysk.core.datetime.datetime import DateTime, timezone, ZoneInfo

from everysk.sdk.entities.query import Query
from everysk.sdk.entities.base import BaseEntity
from everysk.sdk.entities.portfolio.base import Portfolio
from everysk.sdk.entities.portfolio.securities import Securities

from everysk.sdk.base import _mount_self_obj

###############################################################################
#   Base Entity TestCase Implementation
###############################################################################
class TestBaseEntity(TestCase):

    def setUp(self):
        self.expected_response = {
            'filters': [],
            'order': [],
            'projection': None,
            'limit': None,
            'offset': None,
            'page_size': None,
            'page_token': None,
            'distinct_on': []
        }
        self.sample_portfolio_data = {
            'id': 'port_1234567891011211234567890',
            'workspace': 'SampleWorkspace',
            'name': 'SamplePortfolio',
            'tags': ['tag1', 'tag2'],
            'link_uid': 'ABC',
            'description': 'Description',
            'nlv': 1000.0,
            'base_currency': 'USD',
            'date': DateTime.fromisoformat('20231122'),
            'securities': Securities([{'symbol': 'AAPL'}]),
            'version': '1.0',
            'created_on': DateTime.now(),
            'updated_on': DateTime.now(),
            'level': 'HIGH',
            'outstanding_shares': 1000,
            'source_hash': 'XYZ',
            'status': 'OK',
            'portfolio_uid': 'UID123',
            'check_securities': False
        }
        self.portfolio = Portfolio(**self.sample_portfolio_data)

    def test_entity_to_query(self):
        _query = self.portfolio._entity_to_query() # pylint: disable=protected-access
        self.assertDictEqual(_query,
        {
            'filters': [('workspace', '=', 'SampleWorkspace'),
            ('link_uid', '=',
            'ABC'),
                ('name', '<', 'sampleportfolip'),
                ('name', '>=', 'sampleportfolio'),
                ('date', '=', DateTime(2023, 11, 22, 0, 0, tzinfo=ZoneInfo(key='UTC'))),
                ('tags', '=', 'tag1'),
                ('tags', '=', 'tag2')],
            'order': [],
            'projection': None,
            'limit': None,
            'offset': None,
            'page_size': None,
            'page_token': None,
            'distinct_on': []
        })

    def test_get_id_prefix(self):
        with self.assertRaises(NotImplementedError):
            BaseEntity.get_id_prefix()

    def test_generate_id(self):
        with mock.patch.object(BaseEntity, 'get_id_prefix', return_value='P'), \
                mock.patch('everysk.sdk.entities.base.generate_random_id', return_value='12345'):
            self.assertEqual(BaseEntity.generate_id(), 'P12345')

    def test_validate_id(self):
        with mock.patch.object(BaseEntity, '__init__', side_effect=Exception("Mocked Exception")):
            self.assertFalse(BaseEntity.validate_id('invalid_id'))

        with mock.patch.object(BaseEntity, '__init__', return_value=None):
            self.assertTrue(BaseEntity.validate_id('valid_id'))

        with mock.patch.object(BaseEntity, '__init__', return_value=None):
            self.assertFalse(BaseEntity.validate_id(''))

        with mock.patch.object(BaseEntity, '__init__', return_value=None):
            self.assertFalse(BaseEntity.validate_id(None))

    def test_validate(self):
        self.assertTrue(BaseEntity(id="my_id", workspace="my_workspace", name="my_name", created_on=DateTime.now(), updated_on=DateTime.now()).validate())

    def test_raises_validate(self):
        base_entity = BaseEntity()
        self.assertRaisesRegex(
            RequiredError,
            'The created_on attribute is required.',
            base_entity.validate
        )

    def test_get_query(self):
        """ Test _get_query method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities') # pylint: disable=protected-access
        self.assertDictEqual(result, expected_response)

    def test_query(self):
        """ Test query method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities')
        self.assertDictEqual(result, expected_response)

    def test_where(self):
        """ Test where method """
        expected_response = self.expected_response.copy()
        expected_response['filters'] = [('date', '=', DateTime(2023, 8, 10, 12, 0, tzinfo=timezone.utc))]
        result = Portfolio.query.where('date', '=', '2023-08-10')
        self.assertDictEqual(result, expected_response)

    def test_sort_by(self):
        """ Test sort_by method """
        expected_response = self.expected_response.copy()
        expected_response['order'] = ['date']
        result = Portfolio.query.sort_by('date')
        self.assertDictEqual(result, expected_response)

    def test_set_projection(self):
        """ Test set_projection method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities')
        self.assertDictEqual(result, expected_response)

    def test_set_limit(self):
        """ Test set_limit method """
        expected_response = self.expected_response.copy()
        expected_response['limit'] = 5
        result = Portfolio.query.set_limit(5)
        self.assertDictEqual(result, expected_response)

    def test_set_offset(self):
        """ Test set_offset method """
        expected_response = self.expected_response.copy()
        expected_response['offset'] = 10
        result = Portfolio.query.set_offset(10)
        self.assertDictEqual(result, expected_response)

    def test_set_page_size(self):
        """ Test set_page_size method """
        expected_response = self.expected_response.copy()
        expected_response['page_size'] = 15
        result = Portfolio.query.set_page_size(15)
        self.assertDictEqual(result, expected_response)

    def test_set_page_token(self):
        """ Test set_page_token method """
        expected_response = self.expected_response.copy()
        expected_response['page_token'] = 'my_test_page'
        result = Portfolio.query.set_page_token('my_test_page')
        self.assertDictEqual(result, expected_response)

    def test_load(self):
        """ Test find method """
        expected_response = Query(Portfolio)
        expected_response.update(self.expected_response)
        expected_response['filters'] = [('workspace', '=', 'my_workspace')]

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            result = Portfolio.query.where('workspace', 'my_workspace').load(offset=5)

        mock_http_connection.assert_any_call(class_name='Query', params={'offset': 5}, method_name='load', self_obj=_mount_self_obj(expected_response))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsInstance(result, Portfolio)

    def test_list(self):
        """ Test list method """
        expected_response = Query(Portfolio)
        expected_response.update(self.expected_response)
        expected_response['filters'] = [('workspace', '=', 'my_workspace')]

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            Portfolio.query.where('workspace', 'my_workspace').loads(limit=10, offset=5)

        mock_http_connection.assert_any_call(class_name='Query', params={'limit': 10, 'offset': 5}, method_name='loads', self_obj=_mount_self_obj(expected_response))
        mock_http_connection.return_value.get_response.assert_called_once_with()

    def test_page(self):
        """ Test page method """
        expected_response = Query(Portfolio)
        expected_response.update(self.expected_response)
        expected_response['filters'] = [('workspace', '=', 'my_workspace')]

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            Portfolio.query.where('workspace', 'my_workspace').page(page_size=10, page_token='page_token')

        mock_http_connection.assert_any_call(class_name='Query', params={'page_size': 10, 'page_token': 'page_token'}, method_name='page', self_obj=_mount_self_obj(expected_response))
        mock_http_connection.return_value.get_response.assert_called_once_with()

    def test_pages(self):
        """ Test pages method """
        expected_response = Query(Portfolio)
        expected_response.update(self.expected_response)
        expected_response['filters'] = [('workspace', '=', 'my_workspace')]

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            result = Portfolio.query.where('workspace', 'my_workspace').pages(page_size=10)
            next(result)

        mock_http_connection.assert_any_call(class_name='Query', params={'page_size': 10, 'page_token': Undefined}, method_name='page', self_obj=_mount_self_obj(expected_response))
        mock_http_connection.return_value.get_response.assert_called_once_with()

    def test_script_query(self):
        """ Test script_query method """
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            Portfolio.script.fetch(user_input='port_12456', variant='select', workspace=None)

        mock_http_connection.assert_any_call(self_obj={'_klass': 'Portfolio'}, params={'user_input': 'port_12456', 'variant': 'select', 'workspace': None}, class_name='Script', method_name='fetch')
        mock_http_connection.return_value.get_response.assert_called_once_with()

    def test_retrieve(self):
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_portfolio_data
            result = Portfolio.retrieve(entity_id=self.sample_portfolio_data['id'])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id']}, method_name='retrieve', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertEqual(result, self.portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_retrieve_with_none(self):
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = None
            result = Portfolio.retrieve(entity_id=self.sample_portfolio_data['id'])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id']}, method_name='retrieve', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsNone(result)

    def test_create(self):
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_portfolio_data
            result = Portfolio.create(entity_dict=self.sample_portfolio_data)

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_dict': self.sample_portfolio_data}, method_name='create', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertEqual(result, self.portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_modify(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        modified_portfolio = Portfolio(**sample_portfolio_data)
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = sample_portfolio_data
            result = Portfolio.modify(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id'], 'overwrites': {'base_currency': 'BRL'}}, method_name='modify', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertEqual(result, modified_portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_modify_return_none(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = None
            result = Portfolio.modify(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id'], 'overwrites': {'base_currency': 'BRL'}}, method_name='modify', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsNone(result)

    def test_remove(self):
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_portfolio_data
            result = Portfolio.remove(entity_id=self.sample_portfolio_data['id'])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id']}, method_name='remove', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertEqual(result, self.portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_remove_return_none(self):
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = None
            result = Portfolio.remove(entity_id=self.sample_portfolio_data['id'])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id']}, method_name='remove', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsNone(result)

    def test_clone(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        modified_portfolio = Portfolio(**sample_portfolio_data)
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = sample_portfolio_data
            result = Portfolio.clone(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id'], 'overwrites': {'base_currency': 'BRL'}}, method_name='clone', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertEqual(result, modified_portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_clone_return_none(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = None
            result = Portfolio.clone(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id': self.sample_portfolio_data['id'], 'overwrites': {'base_currency': 'BRL'}}, method_name='clone', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsNone(result)

    def test_create_many(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        sample_portfolio_data['check_securities'] = True

        res_modified_portfolio = sample_portfolio_data.copy()
        res_modified_portfolio['check_securities'] = False

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = [self.sample_portfolio_data, res_modified_portfolio]
            result = Portfolio.create_many(entity_dict_list=[self.sample_portfolio_data, sample_portfolio_data])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_dict_list': [self.sample_portfolio_data, sample_portfolio_data]}, method_name='create_many', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsInstance(result, list)
        self.assertEqual(result[0], self.portfolio)
        self.assertEqual(result[1], res_modified_portfolio)
        self.assertIsInstance(result[0], Portfolio)
        self.assertIsInstance(result[1], Portfolio)

    def test_modify_many(self):
        original_portfolio_data = self.sample_portfolio_data.copy()
        original_portfolio_data['name'] = 'modified_1'
        original_portfolio = Portfolio(**original_portfolio_data)

        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['name'] = 'modified_2'
        modified_portfolio = Portfolio(**sample_portfolio_data)

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = [original_portfolio_data, sample_portfolio_data]
            result = Portfolio.modify_many(entity_id_list=[self.sample_portfolio_data['id'], sample_portfolio_data['id']], overwrites=[{'name': 'modified_1'}, {'name': 'modified_2'}])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id_list': [self.sample_portfolio_data['id'], modified_portfolio['id']], 'overwrites': [{'name': 'modified_1'}, {'name': 'modified_2'}]}, method_name='modify_many', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsInstance(result, list)
        self.assertEqual(result[0], original_portfolio)
        self.assertEqual(result[1], modified_portfolio)
        self.assertIsInstance(result[0], Portfolio)
        self.assertIsInstance(result[1], Portfolio)

    def test_remove_many(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['base_currency'] = 'BRL'

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = [self.sample_portfolio_data['id'], sample_portfolio_data['id']]
            result = Portfolio.remove_many(entity_id_list=[self.sample_portfolio_data['id'], sample_portfolio_data['id']])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id_list': [self.sample_portfolio_data['id'], sample_portfolio_data['id']]}, method_name='remove_many', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertListEqual(result, mock_http_connection.return_value.get_response.return_value)

    def test_clone_many(self):
        original_portfolio_data = self.sample_portfolio_data.copy()
        original_portfolio_data['name'] = 'modified_1'
        original_portfolio_data['workspace'] = 'modified_1'
        original_portfolio = Portfolio(**original_portfolio_data)

        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['name'] = 'modified_2'
        sample_portfolio_data['workspace'] = 'modified_2'
        modified_portfolio = Portfolio(**sample_portfolio_data)

        entity_id_list = [original_portfolio['id'], sample_portfolio_data['id']]

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = [original_portfolio_data, sample_portfolio_data]
            result = Portfolio.clone_many(entity_id_list=entity_id_list, overwrites=[{'name': 'modified_1'}, {'name': 'modified_1'}])

        mock_http_connection.assert_any_call(class_name='Portfolio', params={'entity_id_list': entity_id_list, 'overwrites': [{'name': 'modified_1'}, {'name': 'modified_1'}]}, method_name='clone_many', self_obj=None)
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsInstance(result, list)
        self.assertEqual(result[0], original_portfolio)
        self.assertEqual(result[1], modified_portfolio)
        self.assertIsInstance(result[0], Portfolio)
        self.assertIsInstance(result[1], Portfolio)

    def test_save(self):
        portfolio = Portfolio(**self.sample_portfolio_data)
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_portfolio_data
            result = portfolio.save()

        mock_http_connection.assert_any_call(class_name='Portfolio', params={}, method_name='save', self_obj=_mount_self_obj(portfolio))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertEqual(result, self.portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_delete(self):
        portfolio = Portfolio(**self.sample_portfolio_data)
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_portfolio_data
            result = portfolio.delete()

        mock_http_connection.assert_any_call(class_name='Portfolio', params={}, method_name='delete', self_obj=_mount_self_obj(portfolio))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertEqual(result, self.portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_delete_return_none(self):
        portfolio = Portfolio(**self.sample_portfolio_data)
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = None
            result = portfolio.delete()

        mock_http_connection.assert_any_call(class_name='Portfolio', params={}, method_name='delete', self_obj=_mount_self_obj(portfolio))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertIsNone(result)



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.fields import StrField, ListField, IntField, RegexField

###############################################################################
#   Settings Implementation
###############################################################################
WORKER_TEMPLATE_ID_PREFIX = StrField(default='wrkt_', readonly=True)
WORKER_TEMPLATE_DEFAULT_TYPE = StrField(default='BASIC', readonly=True)
WORKER_TEMPLATE_TYPES = ListField(default=['STARTER', 'BASIC', 'FORKER', 'BARRIER', 'CONDITIONAL', 'TRY', 'CATCH', 'ENDER'], readonly=True)

WORKER_TEMPLATE_CATEGORIES = ListField(
    default=[
        'portfolio',
        'parsers',
        'datastore',
        'datastore_generators',
        'report',
        'custom_index',
        'file',
        'OCR',
        'compliance',
        'flow_control',
        'calculation',
        'network',
        'signals',
        'trades',
        'connectors',
        'private_security',
        'miscellaneous'
    ],
    readonly=True
)

WORKER_TEMPLATE_ICONS = ListField(
    default=[
        'analysis',
        'attribution',
        'bridge',
        'bring',
        'brokerage_inserter',
        'calc',
        'calendar',
        'chart',
        'checklist',
        'compliance',
        'conditional',
        'data',
        'default',
        'delete',
        'developers',
        'distribuition',
        'envelope',
        'exposure',
        'fee',
        'file',
        'generator',
        'graphDown',
        'graphUp',
        'guideline',
        'historic',
        'http',
        'image',
        'info',
        'live',
        'look',
        'mapping',
        'matrix',
        'merge',
        'money',
        'notification',
        'optimize',
        'portfolio',
        'positions',
        'property',
        'provision',
        'ranking',
        'retriever',
        'scale',
        'search',
        'sensitivities',
        'sensitivity',
        'settle_provisions',
        'sftp',
        'stress',
        'table',
        'target',
        'text',
        'time',
        'title',
        'trade_allocation',
        'trade_blotter',
        'trade_inserter',
        'trade_position_provision',
        'upgrade',
        'variable',
        'wall',
        'waves',
        'whatIf',
        'window',
        'zip'
    ],
    readonly=True
)

WORKER_TEMPLATE_SCRIPT_SOURCE_MIN_SIZE = IntField(default=1, readonly=True)
WORKER_TEMPLATE_SCRIPT_SOURCE_MAX_SIZE = IntField(default=500000, readonly=True)

WORKER_TEMPLATE_DEFAULT_SCRIPT_RUNTIME = StrField(default='python', readonly=True)
WORKER_TEMPLATE_SCRIPT_RUNTIMES = ListField(default=['python311', 'python', 'ecmascript5'], readonly=True)

WORKER_TEMPLATE_SCRIPT_OUTPUTS_MAX_SIZE = IntField(default=100, readonly=True)
WORKER_TEMPLATE_SCRIPT_INPUTS_MAX_SIZE = IntField(default=200, readonly=True)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock

from everysk.config import settings
from everysk.core.exceptions import FieldValueError
from everysk.core.datetime import DateTime

from everysk.sdk.entities.custom_index.base import CustomIndex
from everysk.sdk.base import _mount_self_obj

###############################################################################
#   Custom Index TestCase Implementation
###############################################################################
class CustomIndexTestCase(TestCase):

    def setUp(self):
        self.sample_data = {
            'symbol': 'CUSTOM:INDEX',
            'name': 'SampleCustomIndex',
            'tags': ['tag1', 'tag2'],
            'description': 'Description',
            'currency': 'USD',
            'data': [[1, 2, 3], [4, 5, 6]],
            'periodicity': 'M',
            'data_type': 'PRICE',
            'base_price': 100,
            'version': '1.0',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
        }
        self.custom_index = CustomIndex(**self.sample_data)

    def test_init(self):
        self.assertEqual(self.custom_index.symbol, self.custom_index.id)
        self.assertEqual(self.custom_index.symbol, self.sample_data['symbol'])
        self.assertEqual(self.custom_index.name, self.sample_data['name'])
        self.assertEqual(self.custom_index.tags, self.sample_data['tags'])
        self.assertEqual(self.custom_index.description, self.sample_data['description'])
        self.assertEqual(self.custom_index.currency, self.sample_data['currency'])
        self.assertEqual(self.custom_index.data, self.sample_data['data'])
        self.assertEqual(self.custom_index.periodicity, self.sample_data['periodicity'])
        self.assertEqual(self.custom_index.data_type, self.sample_data['data_type'])
        self.assertEqual(self.custom_index.base_price, self.sample_data['base_price'])
        self.assertEqual(self.custom_index.version, self.sample_data['version'])
        self.assertEqual(self.custom_index.created_on, self.sample_data['created_on'])
        self.assertEqual(self.custom_index.updated_on, self.sample_data['updated_on'])

    def test_property_id(self):
        self.assertEqual(self.custom_index.id, self.custom_index.symbol)

    def test_setter_id(self):
        self.custom_index.id = 'CUSTOM:NEW_ID'
        self.assertEqual(self.custom_index.id, 'CUSTOM:NEW_ID')
        self.assertEqual(self.custom_index.symbol, 'CUSTOM:NEW_ID')
        self.custom_index.symbol = 'CUSTOM:NEW_ID_2'
        self.assertEqual(self.custom_index.id, 'CUSTOM:NEW_ID_2')
        self.assertEqual(self.custom_index.symbol, 'CUSTOM:NEW_ID_2')

    def test_setter_id_with_invalid_value(self):
        with self.assertRaises(FieldValueError) as e:
            self.custom_index.id = 'Potato'
        self.assertEqual("The value 'Potato' for field 'symbol' must match with this regex: ^CUSTOM:[A-Z0-9_]*$.", e.exception.msg)

        with self.assertRaises(FieldValueError) as e:
            self.custom_index.symbol = 'Potato'
        self.assertEqual("The value 'Potato' for field 'symbol' must match with this regex: ^CUSTOM:[A-Z0-9_]*$.", e.exception.msg)

    def test_get_id_prefix(self):
        self.assertEqual(CustomIndex.get_id_prefix(), settings.CUSTOM_INDEX_SYMBOL_PREFIX)

    def test_validate(self):
        custom_index: CustomIndex = self.custom_index
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = True
            result = custom_index.validate()

        mock_http_connection.assert_called_once_with(class_name='CustomIndex', params={}, method_name='validate', self_obj=_mount_self_obj(custom_index))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertTrue(result)

    def test_entity_to_query(self):
        custom_index: CustomIndex = self.custom_index.copy()
        query = custom_index._entity_to_query() #pylint: disable=protected-access
        self.assertListEqual(query.filters, [
            ('name', '<', 'samplecustomindey'),
            ('name', '>=', 'samplecustomindex'),
            ('tags', '=', 'tag1'),
            ('tags', '=', 'tag2')
        ])
        self.assertEqual(query.order, [])
        self.assertEqual(query.projection, None)
        self.assertEqual(query.limit, None)
        self.assertEqual(query.offset, None)
        self.assertEqual(query.page_size, None)
        self.assertEqual(query.page_token, None)
        self.assertEqual(query.distinct_on, [])

    def test_modify_many(self):
        with self.assertRaises(NotImplementedError) as e:
            CustomIndex.modify_many(entity_id_list=['CUSTOM:INDEX'], overwrites={})

    def test_clone_many(self):
        with self.assertRaises(NotImplementedError) as e:
            CustomIndex.clone_many(entity_id_list=['CUSTOM:INDEX'], overwrites={})



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock

from everysk.config import settings
from everysk.core.datetime import DateTime
from everysk.core.exceptions import RequiredError

from everysk.sdk.entities.file.base import File

from everysk.sdk.base import _mount_self_obj

###############################################################################
#   File TestCase Implementation
###############################################################################
class FileTestCase(TestCase):

    def setUp(self):
        self.sample_data = {
            'id': 'file_1234567891011211234567890',
            'name': 'SampleFile',
            'tags': ['tag1', 'tag2'],
            'description': 'Description',
            'link_uid': 'link_uid',
            'workspace': 'main',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'data': 'base64data',
            'url': '/1234567891011211234567890',
            'content_type': 'text/csv',
            'version': '1.0',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'hash': '4a0ed08522000734bb845f5d72673d9d113c8236',
        }
        self.file = File(**self.sample_data)

    def test_get_id_prefix(self):
        self.assertEqual(File.get_id_prefix(), settings.FILE_ID_PREFIX)

    def test_validate(self):
        file: File = self.file.copy()
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = True
            result = file.validate()

        mock_http_connection.assert_called_once_with(class_name='File', params={}, method_name='validate', self_obj=_mount_self_obj(file))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertTrue(result)

    def test_entity_to_query(self):
        file: File = self.file.copy()
        query = file._entity_to_query() #pylint: disable=protected-access
        self.assertListEqual(query.filters, [
            ('workspace', '=', 'main'),
            ('link_uid', '=', 'link_uid'),
            ('name', '<', 'samplefilf'),
            ('name', '>=', 'samplefile'),
            ('date', '=', self.sample_data['date']),
            ('tags', '=', 'tag1'),
            ('tags', '=', 'tag2'),
            ('url', '=', '/1234567891011211234567890')
        ])
        self.assertEqual(query.order, [])
        self.assertEqual(query.projection, None)
        self.assertEqual(query.limit, None)
        self.assertEqual(query.offset, None)
        self.assertEqual(query.page_size, None)
        self.assertEqual(query.page_token, None)
        self.assertEqual(query.distinct_on, [])

    def test_to_dict(self):
        file: File = self.file.copy()
        file_dict = file.to_dict()
        self.assertEqual(file_dict, {
            'version': '1.0',
            'id': 'file_1234567891011211234567890',
            'name': 'SampleFile',
            'description': 'Description',
            'tags': ['tag1', 'tag2'],
            'link_uid': 'link_uid',
            'workspace': 'main',
            'date': '20230909',
            'data': 'base64data',
            'url': '/file/1234567891011211234567890',
            'content_type': 'text/csv',
            'date_time': '20230909 09:09:09',
            'hash': '4a0ed08522000734bb845f5d72673d9d113c8236',
            'created': 1694250549,
            'updated': 1694250549
        })

    def test_generate_url(self):
        file: File = self.file.copy()
        url = file.generate_url()
        with mock.patch('everysk.sdk.entities.file.base.generate_unique_id') as mock_generate_unique_id:
            mock_generate_unique_id.return_value = '1234567891011211234567890'
            url = file.generate_url()
        self.assertEqual(url, '/1234567891011211234567890')

    def test_required_fields(self):
        file: File = self.file.copy()
        file.id = None
        self.assertRaisesRegex(
            RequiredError,
            'The id attribute is required.',
            file.validate_required_fields
        )

        file.id = 'file_1234567891011211234567890'
        file.name = None
        self.assertRaisesRegex(
            RequiredError,
            'The name attribute is required.',
            file.validate_required_fields
        )

        file.name = 'SampleFile'
        file.workspace = None
        self.assertRaisesRegex(
            RequiredError,
            'The workspace attribute is required.',
            file.validate_required_fields
        )

        file.workspace = 'main'
        file.date = None
        self.assertRaisesRegex(
            RequiredError,
            'The date attribute is required.',
            file.validate_required_fields
        )

        file.date = DateTime(2023, 9, 9, 9, 9, 9, 9)
        file.hash = None
        self.assertRaisesRegex(
            RequiredError,
            'The hash attribute is required.',
            file.validate_required_fields
        )

        file.hash = '4a0ed08522000734bb845f5d72673d9d113c8236'
        file.data = None
        self.assertRaisesRegex(
            RequiredError,
            'The data attribute is required.',
            file.validate_required_fields
        )

        file.data = 'base64data'
        file.content_type = None
        self.assertRaisesRegex(
            RequiredError,
            'The content_type attribute is required.',
            file.validate_required_fields
        )

        file.content_type = 'text/csv'
        file.url = None
        self.assertRaisesRegex(
            RequiredError,
            'The url attribute is required.',
            file.validate_required_fields
        )
